#!/usr/bin/env python
from itertools import takewhile
from math import sqrt, floor
from tempfile import TemporaryFile
from subprocess import Popen, PIPE

# Prime generator using sieve of eratosthenes
# See https://code.activestate.com/recipes/117119-sieve-of-eratosthenes/
def prime_sieve():
	D = {}
	q = 2

	while True:
		if q not in D:
			yield q
			D[q*q] = [q]
		else:
			for p in D[q]:
				D.setdefault(p+q, []).append(p)
			del D[q]
		q += 1

# Get primes smaller than n
def factorbase(n) :
	primes = takewhile(lambda x: x < n, prime_sieve())
	return list(primes)

# Check if x can be factored over F
def check_smoothness(x, F):
	i = 0
	length = len(F)
	while i < length and x > 1 and F[i] <= x:
		if x%F[i] == 0:
			x = x/F[i]
		else:
			i += 1
	if x == 1:
		return True
	else:
		return False

# Generate x such that if x^2 = y mod N then y is smooth over F
def generate_smooths(N, F):
	for j in range(2,1000):
		for k in range(2, j):
			num = floor(sqrt(k*N) + j)
			if check_smoothness(num**2%N, F):
				yield int(num)
			num = floor(sqrt(j*N) +k)
			if check_smoothness(num**2%N, F):
				yield int(num)

def build_system(L, N, F):
	matrix = ([],[],[])
	gen = generate_smooths(N, F)
	
	length = 0
	while length < L:
		x0 = gen.next()
		x = (x0)**2%N
		i = 0
		l = len(F)
		row = [0]*l
		while x > 1 and i < l:
			if x%F[i] == 0:
				row[i] += 1
				x = x/F[i]
			else:
				i += 1

		#row_m2 = map(lambda x: x%2, row)
		row_m2 = []
		for r in row:
			if r != 0:
				row_m2.append(r%2)
			else:
				row_m2.append(0)

		if x == 1 and row_m2 not in matrix[1]:
			matrix[0].append(row)
			matrix[1].append(row_m2)
			matrix[2].append(x0)
			length += 1

	return matrix

def solve_system(system, F):
	out = open('./infile.dat', 'w')
	out.write('%d %d\n' % (len(system[1]), len(F)))
	for row in system[1]:
		out.write(' '.join(map(str, row)))
		out.write('\n')
	out.close()
	
	p = Popen(["./gb", "infile.dat", "outfile.dat"], stdout=PIPE)
	p.wait()

	in_data = open('./outfile.dat', 'r')
	in_data.readline()
	solutions = []
	for row in in_data:
		solutions.append(map(int, row.strip().split(' ')))

	return solutions

def gcd(a,b):
	while b: 
		a,b = b, a%b
	return a

def main():
	B = 10000
	F = factorbase(B)
	L = len(F)+2
	#N = 392742364277
	N = 112120391182534608975671

	system = build_system(L, N, F)
	solutions = solve_system(system, F)

	for solution in solutions:
		x = 1
		y = 1
		i = 0
		exponents = [0]*len(F)


		length = len(solution)
		while i < length:
			if solution[i] == 1:
				x = x*(system[2][i])

				j = 0
				l = len(system[0][i])
				while j < l:
					if system[0][i][j] != 0:
						exponents[j] += system[0][i][j]
						#print "Factor is %d, exp is %d" % (F[j], system[0][i][j])
					j += 1
			i += 1

		k = 0
		m = len(F)
		while k < m:
			#print "k=%d, F[k]=%d" % (k, F[k])
			if exponents[k]%2 != 0:
				print "Exponent not divisible by 2 (factor=%d, exp=%d)" % (F[k], exponents[k])

			y = y*(F[k]**(exponents[k]/2))
			#print "y=%d (factor=%d, exp=%d)" % (y, F[k], exponents[k])
			k += 1


		x = x%N
		y = y%N
		print "Found (x=%d, y=%d)" % (x, y)
		if x**2%N != y**2%N:
			print "Something is broken... (x=%d, y=%d)" % (x, y)

		cur_gcd = gcd(x-y, N)
		if cur_gcd != 1 and cur_gcd != N:
			p = cur_gcd
			q = N/cur_gcd
			break


	print "%d = %d * %d" % (N, p, q)

if __name__ == '__main__':
	main()
